/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */

import './app.scss';
import React, { useState, useEffect, useCallback, useMemo, useRef } from 'react';
import { i18n } from '@kbn/i18n';
import type { TimeRange } from '@kbn/es-query';
import { EuiBreadcrumb, EuiConfirmModal } from '@elastic/eui';
import { useExecutionContext, useKibana } from '@kbn/kibana-react-plugin/public';
import { OnSaveProps } from '@kbn/saved-objects-plugin/public';
import type { VisualizeFieldContext } from '@kbn/ui-actions-plugin/public';
import type { LensAppLocatorParams } from '../../common/locator/locator';
import { LensAppProps, LensAppServices } from './types';
import { LensTopNavMenu } from './lens_top_nav';
import { LensByReferenceInput } from '../embeddable';
import { AddUserMessages, EditorFrameInstance, UserMessagesGetter } from '../types';
import { Document } from '../persistence/saved_object_store';

import {
  setState,
  applyChanges,
  useLensSelector,
  useLensDispatch,
  LensAppState,
  selectSavedObjectFormat,
  updateIndexPatterns,
  selectActiveDatasourceId,
  selectFramePublicAPI,
} from '../state_management';
import { SaveModalContainer, runSaveLensVisualization } from './save_modal_container';
import { LensInspector } from '../lens_inspector_service';
import { getEditPath } from '../../common/constants';
import { isLensEqual } from './lens_document_equality';
import {
  type IndexPatternServiceAPI,
  createIndexPatternService,
} from '../data_views_service/service';
import { replaceIndexpattern } from '../state_management/lens_slice';
import { useApplicationUserMessages } from './get_application_user_messages';

export type SaveProps = Omit<OnSaveProps, 'onTitleDuplicate' | 'newDescription'> & {
  returnToOrigin: boolean;
  dashboardId?: string | null;
  onTitleDuplicate?: OnSaveProps['onTitleDuplicate'];
  newDescription?: string;
  newTags?: string[];
  panelTimeRange?: TimeRange;
};

export function App({
  history,
  onAppLeave,
  redirectTo,
  editorFrame,
  initialInput,
  incomingState,
  redirectToOrigin,
  setHeaderActionMenu,
  datasourceMap,
  visualizationMap,
  contextOriginatingApp,
  topNavMenuEntryGenerators,
  initialContext,
  theme$,
  coreStart,
  savedObjectStore,
}: LensAppProps) {
  const lensAppServices = useKibana<LensAppServices>().services;

  const {
    data,
    dataViews,
    uiActions,
    uiSettings,
    chrome,
    inspector: lensInspector,
    application,
    savedObjectsTagging,
    getOriginatingAppName,
    spaces,
    http,
    notifications,
    executionContext,
    // Temporarily required until the 'by value' paradigm is default.
    dashboardFeatureFlag,
    locator,
    share,
    serverless,
  } = lensAppServices;

  const saveAndExit = useRef<() => void>();

  const dispatch = useLensDispatch();
  const dispatchSetState = useCallback(
    (state: Partial<LensAppState>) => dispatch(setState(state)),
    [dispatch]
  );

  const {
    persistedDoc,
    sharingSavedObjectProps,
    isLinkedToOriginatingApp,
    searchSessionId,
    datasourceStates,
    isLoading,
    isSaveable,
    visualization,
    annotationGroups,
  } = useLensSelector((state) => state.lens);

  const selectorDependencies = useMemo(
    () => ({
      datasourceMap,
      visualizationMap,
      extractFilterReferences: data.query.filterManager.extract.bind(data.query.filterManager),
    }),
    [datasourceMap, visualizationMap, data.query.filterManager]
  );

  const currentDoc = useLensSelector((state) =>
    selectSavedObjectFormat(state, selectorDependencies)
  );

  const shortUrls = useMemo(() => share?.url.shortUrls.get(null), [share]);

  // Used to show a popover that guides the user towards changing the date range when no data is available.
  const [indicateNoData, setIndicateNoData] = useState(false);
  const [isSaveModalVisible, setIsSaveModalVisible] = useState(false);
  const [lastKnownDoc, setLastKnownDoc] = useState<Document | undefined>(undefined);
  const [initialDocFromContext, setInitialDocFromContext] = useState<Document | undefined>(
    undefined
  );
  const [isGoBackToVizEditorModalVisible, setIsGoBackToVizEditorModalVisible] = useState(false);
  const [shouldCloseAndSaveTextBasedQuery, setShouldCloseAndSaveTextBasedQuery] = useState(false);
  const savedObjectId = (initialInput as LensByReferenceInput)?.savedObjectId;

  useEffect(() => {
    if (currentDoc) {
      setLastKnownDoc(currentDoc);
    }
  }, [currentDoc]);

  const showNoDataPopover = useCallback(() => {
    setIndicateNoData(true);
  }, [setIndicateNoData]);

  useExecutionContext(executionContext, {
    type: 'application',
    id: savedObjectId || 'new', // TODO: this doesn't consider when lens is saved by value
    page: 'editor',
  });

  useEffect(() => {
    if (indicateNoData) {
      setIndicateNoData(false);
    }
  }, [setIndicateNoData, indicateNoData, searchSessionId]);

  const getIsByValueMode = useCallback(
    () =>
      Boolean(
        // Temporarily required until the 'by value' paradigm is default.
        dashboardFeatureFlag.allowByValueEmbeddables && isLinkedToOriginatingApp && !savedObjectId
      ),
    [dashboardFeatureFlag.allowByValueEmbeddables, isLinkedToOriginatingApp, savedObjectId]
  );

  useEffect(() => {
    onAppLeave((actions) => {
      if (
        application.capabilities.visualize.save &&
        !isLensEqual(
          persistedDoc,
          lastKnownDoc,
          data.query.filterManager.inject.bind(data.query.filterManager),
          datasourceMap,
          visualizationMap,
          annotationGroups
        ) &&
        (isSaveable || persistedDoc)
      ) {
        return actions.confirm(
          i18n.translate('xpack.lens.app.unsavedWorkMessage', {
            defaultMessage: 'Leave with unsaved changes?',
          }),
          i18n.translate('xpack.lens.app.unsavedWorkTitle', {
            defaultMessage: 'Unsaved changes',
          }),
          undefined,
          i18n.translate('xpack.lens.app.unsavedWorkConfirmBtn', {
            defaultMessage: 'Discard changes',
          }),
          'danger'
        );
      } else {
        return actions.default();
      }
    });
  }, [
    onAppLeave,
    lastKnownDoc,
    isSaveable,
    persistedDoc,
    application.capabilities.visualize.save,
    data.query.filterManager,
    datasourceMap,
    visualizationMap,
    annotationGroups,
  ]);

  const getLegacyUrlConflictCallout = useCallback(() => {
    // This function returns a callout component *if* we have encountered a "legacy URL conflict" scenario
    if (spaces && sharingSavedObjectProps?.outcome === 'conflict' && persistedDoc?.savedObjectId) {
      // We have resolved to one object, but another object has a legacy URL alias associated with this ID/page. We should display a
      // callout with a warning for the user, and provide a way for them to navigate to the other object.
      const currentObjectId = persistedDoc.savedObjectId;
      const otherObjectId = sharingSavedObjectProps?.aliasTargetId!; // This is always defined if outcome === 'conflict'
      const otherObjectPath = http.basePath.prepend(
        `${getEditPath(otherObjectId)}${history.location.search}`
      );
      return spaces.ui.components.getLegacyUrlConflict({
        objectNoun: i18n.translate('xpack.lens.appName', {
          defaultMessage: 'Lens visualization',
        }),
        currentObjectId,
        otherObjectId,
        otherObjectPath,
      });
    }
    return null;
  }, [persistedDoc, sharingSavedObjectProps, spaces, http, history]);

  // Sync Kibana breadcrumbs any time the saved document's title changes
  useEffect(() => {
    const isByValueMode = getIsByValueMode();
    const comesFromVizEditorDashboard =
      initialContext && 'originatingApp' in initialContext && initialContext.originatingApp;
    const breadcrumbs: EuiBreadcrumb[] = [];
    if (
      (isLinkedToOriginatingApp || comesFromVizEditorDashboard) &&
      getOriginatingAppName() &&
      redirectToOrigin
    ) {
      breadcrumbs.push({
        onClick: () => {
          redirectToOrigin();
        },
        text: getOriginatingAppName(),
      });
    }
    if (!isByValueMode) {
      breadcrumbs.push({
        href: application.getUrlForApp('visualize'),
        onClick: (e) => {
          application.navigateToApp('visualize', { path: '/' });
          e.preventDefault();
        },
        text: i18n.translate('xpack.lens.breadcrumbsTitle', {
          defaultMessage: 'Visualize Library',
        }),
      });
    }
    let currentDocTitle = i18n.translate('xpack.lens.breadcrumbsCreate', {
      defaultMessage: 'Create',
    });
    if (persistedDoc) {
      currentDocTitle = isByValueMode
        ? i18n.translate('xpack.lens.breadcrumbsByValue', { defaultMessage: 'Edit visualization' })
        : persistedDoc.title;
    }
    if (
      !persistedDoc?.title &&
      initialContext &&
      'isEmbeddable' in initialContext &&
      initialContext.isEmbeddable
    ) {
      currentDocTitle = i18n.translate('xpack.lens.breadcrumbsEditInLensFromDashboard', {
        defaultMessage: 'Converting {title} visualization',
        values: {
          title: initialContext.title ? `"${initialContext.title}"` : initialContext.visTypeTitle,
        },
      });
    }

    const currentDocBreadcrumb: EuiBreadcrumb = { text: currentDocTitle };
    breadcrumbs.push(currentDocBreadcrumb);
    if (serverless?.setBreadcrumbs) {
      // TODO: https://github.com/elastic/kibana/issues/163488
      // for now, serverless breadcrumbs only set the title,
      // the rest of the breadcrumbs are handled by the serverless navigation
      // the serverless navigation is not yet aware of the byValue/originatingApp context
      serverless.setBreadcrumbs(currentDocBreadcrumb);
    } else {
      chrome.setBreadcrumbs(breadcrumbs);
    }
  }, [
    dashboardFeatureFlag.allowByValueEmbeddables,
    getOriginatingAppName,
    redirectToOrigin,
    getIsByValueMode,
    application,
    chrome,
    isLinkedToOriginatingApp,
    persistedDoc,
    initialContext,
    serverless,
  ]);

  const switchDatasource = useCallback(() => {
    if (saveAndExit && saveAndExit.current) {
      saveAndExit.current();
    }
  }, []);

  const runSave = useCallback(
    (saveProps: SaveProps, options: { saveToLibrary: boolean }) => {
      dispatch(applyChanges());
      return runSaveLensVisualization(
        {
          lastKnownDoc,
          getIsByValueMode,
          savedObjectsTagging,
          initialInput,
          redirectToOrigin,
          persistedDoc,
          onAppLeave,
          redirectTo,
          switchDatasource,
          originatingApp: incomingState?.originatingApp,
          textBasedLanguageSave: shouldCloseAndSaveTextBasedQuery,
          ...lensAppServices,
        },
        saveProps,
        options
      ).then(
        (newState) => {
          if (newState) {
            dispatchSetState(newState);
            setIsSaveModalVisible(false);
            setShouldCloseAndSaveTextBasedQuery(false);
          }
        },
        () => {
          // error is handled inside the modal
          // so ignoring it here
        }
      );
    },
    [
      dispatch,
      lastKnownDoc,
      getIsByValueMode,
      savedObjectsTagging,
      initialInput,
      redirectToOrigin,
      persistedDoc,
      onAppLeave,
      redirectTo,
      switchDatasource,
      incomingState?.originatingApp,
      shouldCloseAndSaveTextBasedQuery,
      lensAppServices,
      dispatchSetState,
    ]
  );

  // keeping the initial doc state created by the context
  useEffect(() => {
    if (lastKnownDoc && !initialDocFromContext) {
      setInitialDocFromContext(lastKnownDoc);
    }
  }, [lastKnownDoc, initialDocFromContext]);

  // if users comes to Lens from the Viz editor, they should have the option to navigate back
  const goBackToOriginatingApp = useCallback(() => {
    if (
      initialContext &&
      'vizEditorOriginatingAppUrl' in initialContext &&
      initialContext.vizEditorOriginatingAppUrl
    ) {
      const [initialDocFromContextUnchanged, currentDocHasBeenSavedInLens] = [
        initialDocFromContext,
        persistedDoc,
      ].map((refDoc) =>
        isLensEqual(
          refDoc,
          lastKnownDoc,
          data.query.filterManager.inject,
          datasourceMap,
          visualizationMap,
          annotationGroups
        )
      );
      if (initialDocFromContextUnchanged || currentDocHasBeenSavedInLens) {
        onAppLeave((actions) => {
          return actions.default();
        });
        application.navigateToApp('visualize', { path: initialContext.vizEditorOriginatingAppUrl });
      } else {
        setIsGoBackToVizEditorModalVisible(true);
      }
    }
  }, [
    annotationGroups,
    application,
    data.query.filterManager.inject,
    datasourceMap,
    initialContext,
    initialDocFromContext,
    lastKnownDoc,
    onAppLeave,
    persistedDoc,
    visualizationMap,
  ]);

  const navigateToVizEditor = useCallback(() => {
    setIsGoBackToVizEditorModalVisible(false);
    if (
      initialContext &&
      'vizEditorOriginatingAppUrl' in initialContext &&
      initialContext.vizEditorOriginatingAppUrl
    ) {
      onAppLeave((actions) => {
        return actions.default();
      });
      application.navigateToApp('visualize', { path: initialContext.vizEditorOriginatingAppUrl });
    }
  }, [application, initialContext, onAppLeave]);

  const initialContextIsEmbedded = useMemo(() => {
    return Boolean(
      initialContext && 'originatingApp' in initialContext && initialContext.originatingApp
    );
  }, [initialContext]);

  const indexPatternService = useMemo(
    () =>
      createIndexPatternService({
        dataViews,
        uiActions,
        core: { http, notifications, uiSettings },
        contextDataViewSpec: (initialContext as VisualizeFieldContext | undefined)?.dataViewSpec,
        updateIndexPatterns: (newIndexPatternsState, options) => {
          dispatch(updateIndexPatterns(newIndexPatternsState));
          if (options?.applyImmediately) {
            dispatch(applyChanges());
          }
        },
        replaceIndexPattern: (newIndexPattern, oldId, options) => {
          dispatch(replaceIndexpattern({ newIndexPattern, oldId }));
          if (options?.applyImmediately) {
            dispatch(applyChanges());
          }
        },
      }),
    [dataViews, uiActions, http, notifications, uiSettings, initialContext, dispatch]
  );

  const onTextBasedSavedAndExit = useCallback(async ({ onSave, onCancel }) => {
    setIsSaveModalVisible(true);
    setShouldCloseAndSaveTextBasedQuery(true);
    saveAndExit.current = () => {
      onSave();
    };
  }, []);

  // remember latest URL based on the configuration
  // url_panel_content has a similar logic
  const shareURLCache = useRef({ params: '', url: '' });

  const shortUrlService = useCallback(
    async (params: LensAppLocatorParams) => {
      const cacheKey = JSON.stringify(params);
      if (shareURLCache.current.params === cacheKey) {
        return shareURLCache.current.url;
      }
      if (locator && shortUrls) {
        // This is a stripped down version of what the share URL plugin is doing
        const shortUrl = await shortUrls.createWithLocator({ locator, params });
        const absoluteShortUrl = await shortUrl.locator.getUrl(shortUrl.params, { absolute: true });
        shareURLCache.current = { params: cacheKey, url: absoluteShortUrl };
        return absoluteShortUrl;
      }
      return '';
    },
    [locator, shortUrls]
  );

  const returnToOriginSwitchLabelForContext =
    initialContext &&
    'isEmbeddable' in initialContext &&
    initialContext.isEmbeddable &&
    !persistedDoc
      ? i18n.translate('xpack.lens.app.replacePanel', {
          defaultMessage: 'Replace panel on {originatingApp}',
          values: {
            originatingApp: initialContext?.originatingApp,
          },
        })
      : undefined;

  const activeDatasourceId = useLensSelector(selectActiveDatasourceId);

  const framePublicAPI = useLensSelector((state) => selectFramePublicAPI(state, datasourceMap));

  const { getUserMessages, addUserMessages } = useApplicationUserMessages({
    coreStart,
    framePublicAPI,
    activeDatasourceId,
    datasourceState:
      activeDatasourceId && datasourceStates[activeDatasourceId]
        ? datasourceStates[activeDatasourceId]
        : null,
    datasource:
      activeDatasourceId && datasourceMap[activeDatasourceId]
        ? datasourceMap[activeDatasourceId]
        : null,
    dispatch,
    visualization: visualization.activeId ? visualizationMap[visualization.activeId] : undefined,
    visualizationType: visualization.activeId,
    visualizationState: visualization,
  });

  return (
    <>
      <div className="lnsApp" data-test-subj="lnsApp" role="main">
        <LensTopNavMenu
          initialInput={initialInput}
          redirectToOrigin={redirectToOrigin}
          getIsByValueMode={getIsByValueMode}
          onAppLeave={onAppLeave}
          runSave={runSave}
          setIsSaveModalVisible={setIsSaveModalVisible}
          setHeaderActionMenu={setHeaderActionMenu}
          indicateNoData={indicateNoData}
          datasourceMap={datasourceMap}
          visualizationMap={visualizationMap}
          title={persistedDoc?.title}
          lensInspector={lensInspector}
          currentDoc={currentDoc}
          isCurrentStateDirty={
            !isLensEqual(
              persistedDoc,
              lastKnownDoc,
              data.query.filterManager.inject.bind(data.query.filterManager),
              datasourceMap,
              visualizationMap,
              annotationGroups
            )
          }
          goBackToOriginatingApp={goBackToOriginatingApp}
          contextOriginatingApp={contextOriginatingApp}
          initialContextIsEmbedded={initialContextIsEmbedded}
          topNavMenuEntryGenerators={topNavMenuEntryGenerators}
          initialContext={initialContext}
          theme$={theme$}
          indexPatternService={indexPatternService}
          onTextBasedSavedAndExit={onTextBasedSavedAndExit}
          getUserMessages={getUserMessages}
          shortUrlService={shortUrlService}
        />
        {getLegacyUrlConflictCallout()}
        {(!isLoading || persistedDoc) && (
          <MemoizedEditorFrameWrapper
            editorFrame={editorFrame}
            showNoDataPopover={showNoDataPopover}
            lensInspector={lensInspector}
            indexPatternService={indexPatternService}
            getUserMessages={getUserMessages}
            addUserMessages={addUserMessages}
          />
        )}
      </div>
      {isSaveModalVisible && (
        <SaveModalContainer
          lensServices={lensAppServices}
          originatingApp={
            isLinkedToOriginatingApp
              ? incomingState?.originatingApp ?? initialContext?.originatingApp
              : undefined
          }
          isSaveable={isSaveable}
          runSave={runSave}
          onClose={() => {
            setIsSaveModalVisible(false);
          }}
          getAppNameFromId={() => getOriginatingAppName()}
          lastKnownDoc={lastKnownDoc}
          onAppLeave={onAppLeave}
          persistedDoc={persistedDoc}
          initialInput={initialInput}
          redirectTo={redirectTo}
          redirectToOrigin={redirectToOrigin}
          initialContext={initialContext}
          returnToOriginSwitchLabel={
            returnToOriginSwitchLabelForContext ??
            (getIsByValueMode() && initialInput
              ? i18n.translate('xpack.lens.app.updatePanel', {
                  defaultMessage: 'Update panel on {originatingAppName}',
                  values: { originatingAppName: getOriginatingAppName() },
                })
              : undefined)
          }
        />
      )}
      {isGoBackToVizEditorModalVisible && (
        <EuiConfirmModal
          maxWidth={600}
          title={i18n.translate('xpack.lens.app.unsavedWorkTitle', {
            defaultMessage: 'Unsaved changes',
          })}
          onCancel={() => setIsGoBackToVizEditorModalVisible(false)}
          onConfirm={navigateToVizEditor}
          cancelButtonText={i18n.translate('xpack.lens.app.goBackModalCancelBtn', {
            defaultMessage: 'Cancel',
          })}
          confirmButtonText={i18n.translate('xpack.lens.app.unsavedWorkConfirmBtn', {
            defaultMessage: 'Discard changes',
          })}
          buttonColor="danger"
          defaultFocusedButton="confirm"
          data-test-subj="lnsApp_discardChangesModalOrigin"
        >
          {i18n.translate('xpack.lens.app.goBackModalMessage', {
            defaultMessage:
              'Your changes here won’t work with your original {contextOriginatingApp} visualization. Leave with unsaved changes and return to {contextOriginatingApp}?',
            values: { contextOriginatingApp },
          })}
        </EuiConfirmModal>
      )}
    </>
  );
}

const MemoizedEditorFrameWrapper = React.memo(function EditorFrameWrapper({
  editorFrame,
  showNoDataPopover,
  getUserMessages,
  addUserMessages,
  lensInspector,
  indexPatternService,
}: {
  editorFrame: EditorFrameInstance;
  lensInspector: LensInspector;
  showNoDataPopover: () => void;
  getUserMessages: UserMessagesGetter;
  addUserMessages: AddUserMessages;
  indexPatternService: IndexPatternServiceAPI;
}) {
  const { EditorFrameContainer } = editorFrame;
  return (
    <EditorFrameContainer
      showNoDataPopover={showNoDataPopover}
      getUserMessages={getUserMessages}
      addUserMessages={addUserMessages}
      lensInspector={lensInspector}
      indexPatternService={indexPatternService}
    />
  );
});
